# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from horizon import exceptions
from horizon import tabs
from horizon.utils import memoized

from openstack_dashboard import api

from bgpvpn_dashboard.api import bgpvpn as bgpvpn_api

from bgpvpn_dashboard.dashboards.project.bgpvpn.network_associations \
    import tables as network_tables


class OverviewTab(tabs.Tab):
    name = _("Overview")
    slug = "overview"
    template_name = "project/bgpvpn/network_associations/_detail_overview.html"

    def get_context_data(self, request):
        network_association = self.tab_group.kwargs['network_association']
        network = self.get_network(network_association.network_id)
        network_association.network_name = network.get('name')
        network_association.network_url = self.get_network_detail_url(
            network_association.network_id)
        return {'network_association': network_association}

    @memoized.memoized_method
    def get_network(self, network_id):
        try:
            network = api.neutron.network_get(self.request, network_id)
        except Exception:
            network = {}
            msg = _('Unable to retrieve network details.')
            exceptions.handle(self.request, msg)

        return network

    @staticmethod
    def get_network_detail_url(network_id):
        return reverse('horizon:project:networks:detail', args=(network_id,))


class NetworkAssociationsTab(tabs.TableTab):
    name = _("Network Associations")
    slug = "network_associations_tab"
    table_classes = (network_tables.NetworkAssociationsTable,)
    template_name = ("horizon/common/_detail_table.html")
    preload = False

    def get_network_associations_data(self):
        try:
            bgpvpn_id = self.tab_group.kwargs['bgpvpn_id']
            network_associations = bgpvpn_api.network_association_list(
                self.request, bgpvpn_id)
            for network_asso in network_associations:
                network = api.neutron.network_get(self.request,
                                                  network_asso.network_id)
                network_asso.network_name = network.name
        except Exception:
            network_associations = []
            msg = _('Network associations list can not be retrieved.')
            exceptions.handle(self.request, msg)
        return network_associations


class NetworkAssociationDetailTabs(tabs.TabGroup):
    slug = "network_association_details"
    tabs = (OverviewTab,)



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import collections
import logging

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from horizon import exceptions
from horizon import forms
from horizon import messages

from openstack_dashboard import api

from bgpvpn_dashboard.api import bgpvpn as bgpvpn_api
from bgpvpn_dashboard.common import bgpvpn as bgpvpn_common

LOG = logging.getLogger(__name__)


class CommonData(forms.SelfHandlingForm):
    fields_order = []
    name = forms.CharField(max_length=255,
                           label=_("Name"),
                           required=False)

    failure_url = reverse_lazy('horizon:project:bgpvpn:index')

    def __init__(self, request, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        if 'keyOrder' in self.fields:
            self.fields.keyOrder = self.fields_order
        else:
            self.fields = collections.OrderedDict(
                (k, self.fields[k]) for k in self.fields_order)

    @staticmethod
    def _del_attributes(attributes, data):
        for attribute in attributes:
            del data[attribute]

    def handle(self, request, data):
        params = {}
        for key in bgpvpn_common.RT_FORMAT_ATTRIBUTES:
            if key in data:
                params[key] = bgpvpn_common.format_rt(data.pop(key, None))
        params.update(data)
        error_msg = _('Something went wrong with BGPVPN %s') % data['name']
        try:
            if self.action == 'update':
                error_msg = _('Failed to update BGPVPN %s') % data['name']
                # attribute tenant_id is required in request when admin user is
                # logged and bgpvpn form from admin menu is used
                if request.user.is_superuser and data.get('tenant_id'):
                    attributes = ('bgpvpn_id', 'type', 'tenant_id')
                # attribute tenant_id does not exist in request
                # when non-admin user is logged
                else:
                    attributes = ('bgpvpn_id', 'type')
                self._del_attributes(attributes, params)
                bgpvpn = bgpvpn_api.bgpvpn_update(request,
                                                  data['bgpvpn_id'],
                                                  **params)
                msg = _('BGPVPN %s was successfully updated.') % data['name']
            elif self.action == 'create':
                error_msg = _('Failed to create BGPVPN %s') % data['name']
                bgpvpn = bgpvpn_api.bgpvpn_create(request, **params)
                msg = _('BGPVPN %s was successfully created.') % data['name']
            else:
                raise Exception(
                    _('Unsupported action type: %s') % self.action)
            LOG.debug(msg)
            messages.success(request, msg)
            return bgpvpn
        except Exception:
            exceptions.handle(request, error_msg, redirect=self.failure_url)
            return False


class EditDataBgpVpn(CommonData):
    bgpvpn_id = forms.CharField(label=_("ID"), widget=forms.TextInput(
        attrs={'readonly': 'readonly'}))
    type = forms.CharField(label=_("Type"), widget=forms.TextInput(
        attrs={'readonly': 'readonly'}))
    fields_order = ['name', 'bgpvpn_id', 'type']

    def __init__(self, request, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        self.action = 'update'


class CreateNetworkAssociation(forms.SelfHandlingForm):
    bgpvpn_id = forms.CharField(widget=forms.HiddenInput())
    network_resource = forms.ChoiceField(
        label=_("Associate Network"),
        widget=forms.ThemableSelectWidget(
            data_attrs=('name', 'id'),
            transform=lambda x: "%s" % x.name_or_id))

    def __init__(self, request, *args, **kwargs):
        super().__init__(
            request, *args, **kwargs)

        # when an admin user uses the project panel BGPVPN, there is no
        # project in data because bgpvpn_get doesn't return it
        project_id = kwargs.get('initial', {}).get("project_id", None)
        if request.user.is_superuser and project_id:
            tenant_id = project_id
        else:
            tenant_id = self.request.user.tenant_id

        try:
            networks = api.neutron.network_list_for_tenant(request, tenant_id)
            if networks:
                choices = [('', _("Choose a network"))] + [(n.id, n) for n in
                                                           networks]
                self.fields['network_resource'].choices = choices
            else:
                self.fields['network_resource'].choices = [('',
                                                            _("No network"))]
        except Exception:
            exceptions.handle(
                request, _("Unable to retrieve networks."))

    def handle(self, request, data):
        try:
            params = self._set_params(data)
            bgpvpn_api.network_association_create(
                request, data['bgpvpn_id'], **params)
            return True
        except Exception:
            exceptions.handle(request, (_('Unable to associate network "%s".')
                                        % data["network_resource"]))
            return False

    def _set_params(self, data):
        params = dict()
        params['network_id'] = data['network_resource']
        return params



# Copyright 2017 <PUT YOUR NAME/COMPANY HERE>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

"""Add tables for port associations

Revision ID: 9a6664f3b8d4
Revises: 0ab4049986b8
Create Date: 2017-06-26 17:34:14.411603

"""

# revision identifiers, used by Alembic.
revision = '9a6664f3b8d4'
down_revision = '0ab4049986b8'


def upgrade():
    op.create_table(
        'bgpvpn_port_associations',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.Column('project_id', sa.String(length=255),
                  index=True, nullable=False),
        sa.Column('bgpvpn_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['bgpvpn_id'], ['bgpvpns.id'],
                                ondelete='CASCADE'),
        sa.Column('port_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.Column('advertise_fixed_ips', sa.Boolean(),
                  server_default=sa.sql.true(), nullable=False),
        sa.UniqueConstraint('bgpvpn_id', 'port_id')
    )

    op.create_table(
        'bgpvpn_port_association_routes',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.Column('port_association_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['port_association_id'],
                                ['bgpvpn_port_associations.id'],
                                ondelete='CASCADE'),
        sa.Column('type', sa.Enum("prefix", "bgpvpn",
                                  name="bgpvpn_port_association_route_type"),
                  nullable=False),
        sa.Column('local_pref', sa.BigInteger(),
                  autoincrement=False, nullable=True),
        # an IPv6 prefix can be up to 49 chars (IPv4-mapped IPv6 string
        # representation: up to 45 chars, plus 4 chars for "/128" which is the
        # highest/longest possible mask)
        sa.Column('prefix', sa.String(49), nullable=True),
        sa.Column('bgpvpn_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['bgpvpn_id'],
                                ['bgpvpns.id'],
                                ondelete='CASCADE')
        # NOTE(tmorin): it would be nice to add some CheckConstraint so that
        # prefix and bgpvpn_id are enforced as NULL unless relevant for the
        # type, and non-NULL when relevant for the type
    )



# Copyright 2016 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine import reflection


"""rename tenant to project

Revision ID: 010308b06b49

Create Date: 2016-06-29 19:42:17.862721

"""

# revision identifiers, used by Alembic.
revision = '23ce05e0a19f'
down_revision = '180baa4183e0'
depends_on = ('0ab4049986b8',)


_INSPECTOR = None


def get_inspector():
    """Reuse inspector"""

    global _INSPECTOR

    if _INSPECTOR:
        return _INSPECTOR

    else:
        bind = op.get_bind()
        _INSPECTOR = reflection.Inspector.from_engine(bind)

    return _INSPECTOR


def get_tables():
    """Returns hardcoded list of tables which have ``tenant_id`` column.

    The list is hard-coded to match the state of the schema when this upgrade
    script is run.
    """

    tables = [
        'bgpvpn_router_associations',
        'bgpvpns',
        'bgpvpn_network_associations',
    ]

    return tables


def get_columns(table):
    """Returns list of columns for given table."""
    inspector = get_inspector()
    return inspector.get_columns(table)


def get_data():
    """Returns combined list of tuples: [(table, column)].

    The list is built from tables with a tenant_id column.
    """

    output = []
    tables = get_tables()
    for table in tables:
        columns = get_columns(table)

        for column in columns:
            if column['name'] == 'tenant_id':
                output.append((table, column))

    return output


def alter_column(table, column):
    old_name = 'tenant_id'
    new_name = 'project_id'

    op.alter_column(
        table_name=table,
        column_name=old_name,
        new_column_name=new_name,
        existing_type=column['type'],
        existing_nullable=column['nullable']
    )


def recreate_index(index, table_name):
    old_name = index['name']
    new_name = old_name.replace('tenant', 'project')

    op.drop_index(op.f(old_name), table_name)
    op.create_index(new_name, table_name, ['project_id'])


def upgrade():
    inspector = get_inspector()

    data = get_data()
    for table, column in data:
        alter_column(table, column)

        indexes = inspector.get_indexes(table)
        for index in indexes:
            if 'tenant_id' in index['name']:
                recreate_index(index, table)


def contract_creation_exceptions():
    """Special migration for the blueprint to support Keystone V3.

    We drop all tenant_id columns and create project_id columns instead.
    """

    return {
        sa.Column: ['.'.join([table, 'project_id']) for table in get_tables()],
        sa.Index: get_tables()
    }



# Copyright 2018 OpenStack Fundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

"""Add standard attributes

Revision ID: 7a9482036ecd
Revises: 666c706fea3b
Create Date: 2018-04-04 10:12:40.399032

"""

# revision identifiers, used by Alembic.
revision = '7a9482036ecd'
down_revision = '666c706fea3b'


def upgrade():
    for table in ('bgpvpns', 'bgpvpn_network_associations',
                  'bgpvpn_router_associations', 'bgpvpn_port_associations'):
        op.add_column(table, sa.Column('standard_attr_id', sa.BigInteger(),
                                       nullable=True))
